; TTN Interrupt Slot Handler driver
; Simon N Goodwin 30th October 2023
; v1.9 Last revised 24th March 2024

limit   EQU  512                ; Size of driver area
slots   EQU  64                 ; Number of slots
stride  EQU  4                  ; Bytes per slot
base    EQU  49160              ; Code assembly address
headlen EQU  8                  ; Driver header size
recount EQU  (length-relocs)/2  ; Number of relocations

; NextREG interface constants

select  EQU  9275               ; NextREG select port
MMU6    EQU  86                 ; Penultimate MMU register
MMU7    EQU  87                 ; MMU register for top 8K

; Lifecycle API codes defined by NextZXOS

getinit EQU  128                ; Initialisation API
getshut EQU  129                ; Shutdown API, unused

; API codes for reading slots (reserved: 0, 65..100, 103..127)

getcode EQU  101                ; Return code address
getdata EQU  102                ; Return data address

; This is the header used to load the driver

        ORG  base-headlen
        DEFB "NDRV"             ; Magic start marker
        DEFB 8+128              ; Includes interrupt flag
        DEFB recount            ; Number of relocations
        DEFB 128                ; Call Init, no DivMMC
        DEFB 0                  ; No additional banks

; Here starts the actual 512 byte driver code and data

        ORG  base,0             ; Zeus ORG 0 DISP base

; This is the command entry point; the command ID is in B

api     JP   comms              ; Interpret DRIVER commands

; This is the frame interrupt handler which calls the slots

intr    EX   AF,AF'             ; Save alternate AF
        PUSH AF
        PUSH IX                 ; Save index registers
        PUSH IY

        EXX                     ; Save alternate register set
        PUSH HL
        PUSH DE
        PUSH BC                 ; 12 bytes stacked

scan    LD   HL,slot1           ; Test code vector high bytes
        LD   B,slots            ;   for this many slots
        LD   DE,stride          ;     at this interval
        XOR  A                  ; Zero means slot is inactive

try     OR   (HL)               ; Check if slot is active
        JR   NZ,use

        ADD  HL,DE              ; Advance to next slot
used    DJNZ try

        POP  BC                 ; Restore alternate regs
        POP  DE
        POP  HL
        EXX                     ; Revert to default set

        POP  IY                 ; Restore index registers
        POP  IX
        POP  AF                 ; Restore alternate AF
        EX   AF,AF'

        RET

; HL -> an active slot; find corresponding code and data

use     PUSH BC                 ; Save slot loop count
        LD   B,(HL)             ; Code vector to BC
        INC  HL
        LD   C,(HL)
        INC  HL
        LD   D,(HL)             ; Data vector to DE
        INC  HL
        LD   E,(HL)
        INC  HL
        PUSH HL                 ; Save address of next slot

        JR   rel1               ; Disable 1.8 bad magic

; Check if SP is below 48K before attempting any paging

        LD   HL,0
        ADD  HL,SP              ; The offical way to read SP
        LD   A,%11000000        ; Bits set in address 49152+
        AND  H                  ; Ignore the lower-order bits
        CP   %11000000
        JR   Z,rel1             ; Never page out the stack!

; If either the code or data address is above 48K we must
; make sure that the normal BASIC bank 0 is mapped there
; while the interrupt handler runs, then restore whatever
; was mapped there afterwards, as NextZXOS may change it.

        LD   H,%11000000        ; Mask top address bits
        LD   A,D                ; Consider data address
        AND  H                  ; Ignore other bits
        CP   H                  ; Are both still set?
        JR   Z,setBank

        LD   A,B                ; Consider code address
        AND  H
        CP   H
        JR   NZ,rel1            ; Nothing needs saving

; Save and restore top 16K bank mapping from REG 86 and 87

setBank PUSH BC                 ; Prepare for port input
        LD   BC,select          ; Address REG select port
        LD   A,MMU6             ; MMU6 maps 48K to 56K
        OUT  (C),A
        INC  B                  ; Advance to REG data port
        IN   L,(C)              ; L = REG 86
        OUT  (C),0              ; Put start of BANK 0 at 48K
        LD   A,MMU7             ; Index MMU7, 56K to 64K
        DEC  B                  ; Back to REG select port
        OUT  (C),A              ; Select MMU7
        INC  B
        IN   H,(C)              ; H = REG 87
        LD   A,1                ; Top 8K of BANK 0 page number
        OUT  (C),A              ; Map the rest of BANK 0 at 56K
        POP  BC                 ; Restore code address
        PUSH HL                 ; Save MMU6 & MMU7 contents
rel0    CALL jpBC

        POP  HL                 ; Recover stored MMU contents
        LD   A,L
        NEXTREG MMU6,A          ; Restore 48K to 56K mapping
        LD   A,H
        NEXTREG MMU7,A          ; Restore 56K to 64K mapping
        JR   proceed

rel1    CALL jpBC

proceed POP  HL                 ; Recover the try loop context
        POP  BC
        LD   DE,stride          ; Reload slot size for stepping
        XOR  A                  ; Reload empty slot marker
        JR   used

jpBC    LD   H,B                ; Slot at BC also starts at HL
        LD   L,C                ; Associated data starts at DE
        JP   (HL)

; Clear first byte of slots at initialisation or DRIVER 8,128

init    LD   B,slots            ; Count all the slots
        LD   DE,stride          ; Interval between entries
rel2    LD   HL,slot1           ; Start at the first slot

empty   LD   (HL),D             ; We know that stride < 256
        ADD  HL,DE
        DJNZ empty

        AND  A                  ; Clear carry = success
        RET

; Commands 1..<slots> initialise the chosen slot
; GETCODE and GETDATA commands read corresponding fields

comms   BIT  7,B                ; Quickly sift out OS IDs
        JR   NZ,system

        LD   A,slots
        CP   B                  ; Is B a slot number, or 0?
        JR   NC,setslot

; Validate parameter in DE, the slot number to read

        LD   A,D
        OR   A
        JR   NZ,error           ; Reject DE > 255

        OR   E
        JR   Z,error            ; Disregard slot 0

        LD   A,slots
        CP   E
        JR   C,error            ; Reject E > slots

        EX   DE,HL              ; Move parameter into HL
        ADD  HL,HL
        ADD  HL,HL              ; Scale parameter by stride
        LD   A,getcode
        CP   B                  ; Check for code-reading ID
        JR   NZ,notcode

rel3    LD   DE,slot1           ; Find slot from offset in HL
        ADD  HL,DE
        LD   B,(HL)             ; Return BC
        INC  HL
        LD   C,(HL)

done    AND  A                  ; Flag success
        RET

notcode LD   A,getdata
        CP   B                  ; Check for data-reading ID
        JR   NZ,error

        INC  HL                 ; Point at data field in slot
        INC  HL
        JR   rel3

system  LD   A,getinit
        CP   B
        JR   Z,init             ; Mark all slots free for use

        LD   A,getshut
        CP   B
        JR   Z,done             ; No specific shutdown action

error   LD   A,128
        ADD  A                  ; Carry flag signals failure
        RET

; DE is the code address for slot B, HL is the data address

setslot LD   A,B
        OR   A
        JR   Z,error            ; There is no slot 0
        DEC  B

        PUSH HL
        LD   H,0                ; Form 16-bit offset
        LD   L,B
        ADD  HL,HL              ; Double twice to fit stride
        ADD  HL,HL              ; Slot offset is in HL

rel4    LD   BC,slot1
        ADD  HL,BC              ; HL -> slot B
        POP  BC                 ; Get data address in BC
        LD   (HL),D             ; Store high byte first
        INC  HL
        LD   (HL),E             ; Store low byte next
        INC  HL
        LD   (HL),B             ; Data address high byte
        INC  HL                 ; Advance to store low byte
        LD   (HL),C
        AND  A                  ; No carry, flag success
        RET
end                             ; Must be <= limit-slots*size!

; Allocate slots up to the limit; pad space with temporary text

        ORG base+limit-(slots*stride),limit-(slots*stride)
slot1   DEFB "Interupt Slot Handler v1.9 2023 Simon N Goodwin"
        DEFB 13,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

; Relocation data follows slots at the end of the driver

        ORG base+limit,limit
relocs  DEFW  api+2
        DEFW scan+2
        DEFW rel0+2
        DEFW rel1+2
        DEFW rel2+2
        DEFW rel3+2
        DEFW rel4+2
length

        SAVE "Slot.drv",base-headlen,length+headlen



